from .utils import cart2pol, denoise, morph
import mahotas as mh
import numpy as np

class PAGE:
    def __init__(self,direction_bins, h=None, w=None):
        self.h = h
        self.w = w
        self.direction_bins = direction_bins
    
    
    def load_img(self, img_file):
        # load the image from the image file
        self.img = mh.imread(img_file)
        # convert to grayscale if it is RGB
        if self.img.ndim == 3:
            self.img = mh.colors.rgb2grey(self.img)
        # read the image size or resize to the indicated size (height x width)
        if not self.h and not self.w:
            self.h = self.img.shape[0]
            self.w = self.img.shape[1]
        else:
            self.img = mh.imresize(self.img, [self.h, self.w])
        
    
    def init_kernel(self, mu_1, mu_2, sigma_1, sigma_2, S1, S2):
        # initialize the filter arrays of page
        # mu_1            Center frequency of a normal/Gaussian distributed passband filter ϕ1
        # mu_2            Center frequency of log-normal distributed distributed passband filter ϕ2
        # sigma_1         Standard deviation sigma of normal/Gaussian distributed passband filter ϕ1
        # sigma_2         Standard deviation sigma of log-normal distributed passband filter ϕ2
        # S1              Strength (Amplitude) of ϕ1 filter
        # S2              Strength (Amplitude) of ϕ2 filter
        
        # set the frequency grid
        u = np.linspace(-0.5, 0.5, self.h)
        v = np.linspace(-0.5, 0.5, self.w)
        [U1, V1] = (np.meshgrid(u, v))
        U = U1.T
        V = V1.T
        [self.THETA, self.RHO] = cart2pol(U, V)

        min_direction = np.pi/180
        direction_span = np.pi / self.direction_bins 
        directions = np.arange(min_direction, np.pi, direction_span)
    
        # create PAGE kernels channel by channel
        self.page_kernel = np.zeros([self.h, self.w, self.direction_bins])
        for i in range(self.direction_bins):
            tetav = directions[i]
            # Project onto new directionality basis for PAGE filter creation
            Uprime = U * np.cos(tetav) + V * np.sin(tetav)
            Vprime = -U * np.sin(tetav) + V * np.cos(tetav)

            # Create Normal component of PAGE filter
            Phi_1 = np.exp(-0.5 * ((abs(Uprime) - mu_1) / sigma_1)** 2) / (1 * np.sqrt(2 * np.pi) * sigma_1)
            Phi_1 = (Phi_1 / np.max(Phi_1[:])) * S1

            # Create Log-Normal component of PAGE filter
            Phi_2 = np.exp(-0.5 * ((np.log(abs(Vprime)) - mu_2) / sigma_2) ** 2) / (abs(Vprime) * np.sqrt(2 * np.pi) * sigma_2)
            Phi_2 = (Phi_2 / np.max(Phi_2[:])) * S2

            # Add overall directional filter to PAGE filter array
            self.page_kernel[:,:,i]= Phi_1 * Phi_2
    

    def apply_kernel(self, sigma_LPF, thresh_min, thresh_max, morph_flag):
        # denoise on the loaded image
        self.img_denoised = denoise(img=self.img, rho=self.RHO, sigma_LPF=sigma_LPF)
        self.page_output = np.zeros([self.h, self.w, self.direction_bins])
        # apply the kernel channel by channel
        for i in range(self.direction_bins):
            self.img_page = np.fft.ifft2((np.fft.fft2(self.img_denoised)) * np.fft.fftshift(np.exp(-1j * self.page_kernel[:,:,i])))
            self.page_feature = np.angle(self.img_page)
            # apply morphological operation if applicable
            if morph_flag == 0:
                self.page_output[:,:,i] = self.page_feature
            else:
                self.page_output[:,:,i] = morph(img=self.img, feature=self.page_feature, thresh_max=thresh_max, thresh_min=thresh_min)


    def create_page_edge(self):
        # Create a weighted color image of PAGE output to visualize directionality of edges
        weight_step = 255*3/self.direction_bins
        color_weight = np.arange(0, 255, weight_step)
        self.page_edge = np.zeros([self.h, self.w, 3])
        step_edge = int(round(self.direction_bins/3))
        for i in range(step_edge):
            self.page_edge[:,:,0] = color_weight[i] * self.page_output[:,:,i] + self.page_edge[:,:,0]
            self.page_edge[:,:,1] = color_weight[i] * self.page_output[:,:,i+step_edge] + self.page_edge[:,:,1]
            self.page_edge[:,:,2] = color_weight[i] * self.page_output[:,:,i+(2*step_edge)] + self.page_edge[:,:,2]

        self.page_edge = (self.page_edge - np.min(self.page_edge)) / (np.max(self.page_edge) - np.min(self.page_edge))


    def run(self, img_file, mu_1, mu_2, sigma_1, sigma_2, S1, S2, sigma_LPF, thresh_min, thresh_max, morph_flag):
        # wrap load_img, init_kernel, apply_kernel, create_page_edge in one run
        self.load_img(img_file)
        self.init_kernel(mu_1, mu_2, sigma_1, sigma_2, S1, S2)
        self.apply_kernel(sigma_LPF, thresh_min, thresh_max, morph_flag)
        self.create_page_edge()
        
        return self.page_edge


from .utils import cart2pol, denoise, morph
import mahotas as mh
import numpy as np

class PST:
    def __init__(self, h=None, w=None):
        self.h = h
        self.w = w
    

    def load_img(self, img_file):
        # load the image from the image file
        self.img = mh.imread(img_file)
        # convert to grayscale if it is RGB
        if self.img.ndim == 3:
            self.img = mh.colors.rgb2grey(self.img)
        # read the image size or resize to the indicated size (height x width)
        if not self.h and not self.w:
            self.h = self.img.shape[0]
            self.w = self.img.shape[1]
        else:
            self.img = mh.imresize(self.img, [self.h, self.w])
        

    def init_kernel(self, phase_strength, warp_strength):
        # set the frequency grid
        u = np.linspace(-0.5, 0.5, self.h)
        v = np.linspace(-0.5, 0.5, self.w)
        [U1, V1] = (np.meshgrid(u, v))
        U = U1.T
        V = V1.T
        [self.THETA, self.RHO] = cart2pol(U, V)
        # construct the PST Kernel
        self.pst_kernel = np.multiply(warp_strength*self.RHO, np.arctan(warp_strength*self.RHO))-0.5*np.log(1+np.power(warp_strength*self.RHO,2))
        self.pst_kernel = self.pst_kernel/np.max(self.pst_kernel)*phase_strength


    def apply_kernel(self, sigma_LPF, thresh_min, thresh_max, morph_flag):
        # denoise on the loaded image
        self.img_denoised = denoise(img=self.img, rho=self.RHO, sigma_LPF=sigma_LPF)
        # apply the pst kernel
        self.img_pst = np.fft.ifft2((np.fft.fft2(self.img_denoised)) * np.fft.fftshift(np.exp(-1j * self.pst_kernel)))
        self.pst_feature = np.angle(self.img_pst)
        # apply morphological operation if applicable
        if morph_flag == 0:
            self.pst_output = self.pst_feature
        else:
            self.pst_output = morph(img=self.img, feature=self.pst_feature, thresh_max=thresh_max, thresh_min=thresh_min)


    def run(self, img_file, phase_strength, warp_strength, sigma_LPF, thresh_min, thresh_max, morph_flag):
        # wrap load_img, init_kernel, apply_kernel in one run
        self.load_img(img_file)
        self.init_kernel(phase_strength, warp_strength)
        self.apply_kernel(sigma_LPF, thresh_min, thresh_max, morph_flag)

        return self.pst_output



from .utils import cart2pol_torch, denoise_torch, morph_torch
import torch
import torchvision
from torchvision.io import read_image
from torchvision.transforms.functional import rgb_to_grayscale, resize
import numpy as np


class PST_GPU:
    def __init__(self, device, h=None, w=None):
        self.h = h
        self.w = w
        self.device = device
    

    def load_img(self, img_file):
        # load the image from the image file
        # torchvision read_image only supports 'jpg' and 'png'
        if img_file.split('.')[-1] in ['jpg', 'png', 'jpeg']:
            self.img = torchvision.io.read_image(img_file).to(self.device)
            # convert to grayscale if it is RGB
            if self.img.dim() == 3 and self.img.shape[0]!=1:
                self.img = rgb_to_grayscale(self.img)
            # read the image size or resize to the indicated size (height x width)
            if not self.h and not self.w:
                self.img = torch.squeeze(self.img)
                self.h = self.img.shape[0]
                self.w = self.img.shape[1]
            else:
                self.img = torch.squeeze(resize(self.img, [self.h, self.w]))
        else:
            # use mahotas to load other format of image
            self.img = mh.imread(img_file)
            if self.img.ndim == 3:
                self.img = mh.colors.rgb2grey(self.img)
            if not self.h and not self.w:
                self.h = self.img.shape[0]
                self.w = self.img.shape[1]
            else:
                self.img = mh.imresize(self.img, [self.h, self.w])
            
            self.img = torch.from_numpy(self.img).to(self.device)


    def init_kernel(self, phase_strength, warp_strength):
        # set the frequency grid
        u = torch.linspace(-0.5, 0.5, self.h, device=self.device).float()
        v = torch.linspace(-0.5, 0.5, self.w, device=self.device).float()
        [U, V] = (torch.meshgrid(u, v))
        [self.THETA, self.RHO] = cart2pol_torch(U, V)
        # construct the PST Kernel
        self.pst_kernel=torch.multiply(warp_strength*self.RHO, torch.arctan(warp_strength*self.RHO))-0.5*torch.log(1+torch.pow(warp_strength*self.RHO,2))
        self.pst_kernel=self.pst_kernel/torch.max(self.pst_kernel)*phase_strength


    def apply_kernel(self, sigma_LPF, thresh_min, thresh_max, morph_flag):
        # denoise on the loaded image
        self.img_denoised = denoise_torch(img=self.img, rho=self.RHO, sigma_LPF=sigma_LPF)
        # apply the pst kernel
        self.img_pst = torch.fft.ifft2((torch.fft.fft2(self.img_denoised)) * torch.fft.fftshift(torch.exp(-1j * self.pst_kernel)))
        self.pst_feature = torch.angle(self.img_pst)
        # apply morphological operation if applicable
        if morph_flag == 0:
            self.pst_output = self.pst_feature
        else:
            kernel = torch.tensor([[0.0, 1.0, 0.0],
                               [1.0, 1.0, 1.0],
                               [0.0, 1.0, 0.0]]).to(self.device)
            self.pst_output = morph_torch(img=self.img, feature=self.pst_feature, thresh_max=thresh_max, thresh_min=thresh_min, kernel=kernel, device=self.device)
    

    def run(self, img_file, phase_strength, warp_strength, sigma_LPF, thresh_min, thresh_max, morph_flag):
        # wrap load_img, init_kernel, apply_kernel in one run
        self.load_img(img_file)
        self.init_kernel(phase_strength, warp_strength)
        self.apply_kernel(sigma_LPF, thresh_min, thresh_max, morph_flag)

        return self.pst_output



from .utils import cart2pol_torch, denoise_torch, morph_torch
import torch
import torchvision
from torchvision.io import read_image
from torchvision.transforms.functional import rgb_to_grayscale, resize
import numpy as np

class PAGE_GPU:
    def __init__(self, direction_bins, device, h=None, w=None):
        self.h = h
        self.w = w
        self.direction_bins = direction_bins
        self.device = device

    def load_img(self, img_file):
        # load the image from the image file
        # torchvision read_image only supports 'jpg' and 'png'
        if img_file.split('.')[-1] in ['jpg', 'png', 'jpeg']:
            self.img = torchvision.io.read_image(img_file).to(self.device)
            # convert to grayscale if it is RGB
            if self.img.dim() == 3 and self.img.shape[0]!=1:
                self.img = rgb_to_grayscale(self.img)
            # read the image size or resize to the indicated size (height x width)
            if not self.h and not self.w:
                self.img = torch.squeeze(self.img)
                self.h = self.img.shape[0]
                self.w = self.img.shape[1]
            else:
                self.img = torch.squeeze(resize(self.img, [self.h, self.w]))
        
        else:
            # use mahotas to load other format of image
            self.img = mh.imread(img_file)
            if self.img.ndim == 3:
                self.img = mh.colors.rgb2grey(self.img)
            if not self.h and not self.w:
                self.h = self.img.shape[0]
                self.w = self.img.shape[1]
            else:
                self.img = mh.imresize(self.img, [self.h, self.w])
            
            self.img = torch.from_numpy(self.img).to(self.device)


    def init_kernel(self, mu_1, mu_2, sigma_1, sigma_2, S1, S2):
        # initialize the filter arrays of page
        # mu_1            Center frequency of a normal/Gaussian distributed passband filter ϕ1
        # mu_2            Center frequency of log-normal distributed distributed passband filter ϕ2
        # sigma_1         Standard deviation sigma of normal/Gaussian distributed passband filter ϕ1
        # sigma_2         Standard deviation sigma of log-normal distributed passband filter ϕ2
        # S1              Strength (Amplitude) of ϕ1 filter
        # S2              Strength (Amplitude) of ϕ2 filter

        # set the frequency grid
        u = torch.linspace(-0.5, 0.5, self.h, device=self.device).float()
        v = torch.linspace(-0.5, 0.5, self.w, device=self.device).float()
        [U, V] = (torch.meshgrid(u, v))
        [self.THETA, self.RHO] = cart2pol_torch(U, V)

        min_direction = np.pi/180
        direction_span = np.pi / self.direction_bins 
        directions = torch.arange(min_direction, np.pi, direction_span)
    
        # create PAGE kernels in parallel by broadcasting
        tetavs = torch.unsqueeze(directions, dim=0).to(self.device)
        Us = torch.unsqueeze(U, dim=-1)
        Vs = torch.unsqueeze(V, dim=-1)
        Uprimes = Us * torch.cos(tetavs) + Vs * torch.sin(tetavs)
        Vprimes = -Us * torch.sin(tetavs) + Vs * torch.cos(tetavs)

        Phi_1s = torch.exp(-0.5 * ((torch.abs(Uprimes) - mu_1) / sigma_1)** 2) / (1 * np.sqrt(2 * np.pi) * sigma_1)
        Phi_1s = (Phi_1s / torch.max(Phi_1s.view(-1, self.direction_bins), dim=0)[0]) * S1

        Phi_2s = torch.exp(-0.5 * ((torch.log(torch.abs(Vprimes)) - mu_2) / sigma_2) ** 2) / (abs(Vprimes) * np.sqrt(2 * np.pi) * sigma_2)
        Phi_2s = (Phi_2s / torch.max(Phi_2s.view(-1, self.direction_bins), dim=0)[0]) * S2
        self.page_kernel = Phi_1s * Phi_2s

    
    def apply_kernel(self, sigma_LPF, thresh_min, thresh_max, morph_flag):
        # denoise on the loaded image
        self.img_denoised = denoise_torch(img=self.img, rho=self.RHO, sigma_LPF=sigma_LPF)
        # apply the page kernel
        self.img_page = torch.fft.ifft2((torch.fft.fft2(self.img_denoised).unsqueeze(-1)) * torch.fft.fftshift(torch.exp(-1j * self.page_kernel), dim=(0,1)), dim=(0,1))
        self.page_feature = torch.angle(self.img_page)
        # apply morphological operation if applicable
        if morph_flag == 0:
            self.page_output = self.page_feature
        else:
            kernel = torch.tensor([[0.0, 1.0, 0.0],
                                   [1.0, 1.0, 1.0],
                                   [0.0, 1.0, 0.0]]).to(self.device)

            self.page_output = morph_torch(img=self.img, feature=self.page_feature, thresh_min=thresh_min, thresh_max=thresh_max, kernel=kernel, device=self.device)


    def create_page_edge(self):
        # Create a weighted color image of PAGE output to visualize directionality of edges
        weight_step = 255*3/self.direction_bins
        color_weight = torch.arange(0, 255, weight_step).to(self.device)
        self.page_edge = torch.zeros([self.h, self.w, 3]).to(self.device)
        step_edge = int(round(self.direction_bins/3))
        for i in range(step_edge):
            self.page_edge[:,:,0] = color_weight[i] * self.page_output[:,:,i] + self.page_edge[:,:,0]
            self.page_edge[:,:,1] = color_weight[i] * self.page_output[:,:,i+step_edge] + self.page_edge[:,:,1]
            self.page_edge[:,:,2] = color_weight[i] * self.page_output[:,:,i+(2*step_edge)] + self.page_edge[:,:,2]
        
        self.page_edge = (self.page_edge - torch.min(self.page_edge)) / (torch.max(self.page_edge) - torch.min(self.page_edge))


    def run(self, img_file, mu_1, mu_2, sigma_1, sigma_2, S1, S2, sigma_LPF, thresh_min, thresh_max, morph_flag):
        # wrap load_img, init_kernel, apply_kernel, create_page_edge in one run
        self.load_img(img_file)
        self.init_kernel(mu_1, mu_2, sigma_1, sigma_2, S1, S2)
        self.apply_kernel(sigma_LPF, thresh_min, thresh_max, morph_flag)
        self.create_page_edge()
        
        return self.page_edge



import numpy as np
import torch
import torch.fft
import mahotas as mh
import kornia

'''
in utils.py, we define common operations in both PST and PAGE

cart2pol: transform cartesian coordinates to polar coordinates
denoise: apply a low pass filter to the original image
morph: morphological operation from analog features to digital features

'''

def cart2pol(x, y):
    '''
    transform cartesian coordinates to polar coordinates
    Arguments: cartesian coordinates: x, y
    Return: polar coordinates: rho, theta

    '''
    theta = np.arctan2(y, x)
    rho = np.hypot(x, y)
    return (theta, rho)

def cart2pol_torch(x,y):
    '''
    cart2pol GPU version

    '''
    theta = torch.atan2(y, x)
    rho = torch.hypot(x, y)
    return (theta, rho)


def denoise(img, rho, sigma_LPF):
    '''
    apply a low pass filter to the original image
    Arguments: original image, polar coordinates rho, sigma of the low-pass-filter
    Return: filtered image

    '''
    img_orig_f = np.fft.fft2(img)
    expo = np.fft.fftshift(np.exp(-0.5*np.power((np.divide(rho, np.sqrt((sigma_LPF**2)/np.log(2)))),2)))
    img_filtered = np.real(np.fft.ifft2((np.multiply(img_orig_f,expo))))

    return img_filtered


def denoise_torch(img, rho, sigma_LPF):
    '''
    denoise GPU version

    '''
    img_orig_f = torch.fft.fft2(img)
    expo = torch.fft.fftshift(torch.exp(-0.5*torch.pow((torch.divide(rho, np.sqrt((sigma_LPF ** 2) / np.log(2)))), 2)))
    img_filtered = torch.real(torch.fft.ifft2((torch.mul(img_orig_f, expo))))

    return img_filtered


def morph(img, feature, thresh_min, thresh_max):
    '''
    apply morphological operation to transform analog features to digial features
    Argument: original image, analog features output by PST/PAGE, threshhold min and threshold max
    Return: digital feature after morphological operations

    '''
    digital_feature = np.zeros(feature.shape)
    
    digital_feature[feature>thresh_max] = 1 # Bi-threshold decision
    digital_feature[feature<thresh_min] = 1 # as the output phase has both positive and negative values
    digital_feature[img<(np.amax(img)/20)]=0 # Removing edges in the very dark areas of the image (noise)

    digital_feature = mh.thin(digital_feature, 1)
    digital_feature = mh.bwperim(digital_feature, 4)
    digital_feature = mh.thin(digital_feature, 1)
    digital_feature = mh.erode(digital_feature, np.ones((1, 1)))

    return digital_feature.astype(np.float32)


def morph_torch(img, feature, thresh_min, thresh_max, kernel, device):
    '''
    morph GPU version (torch and kornia)

    '''
    digital_feature = torch.zeros(feature.shape).to(device)
    digital_feature[feature>thresh_max] = 1 # Bi-threshold decision
    digital_feature[feature<thresh_min] = 1 # as the output phase has both positive and negative values
    digital_feature[img<(torch.max(img)/20)]=0 # Removing edges in the very dark areas of the image (noise)

    # for PST
    if len(feature.shape)==2:
        digital_feature = kornia.utils.image._to_bchw(digital_feature)
        digital_feature = kornia.morphology.closing(digital_feature, kernel)

    # for PAGE
    elif len(feature.shape)==3:
        digital_feature = torch.permute(digital_feature,(2,0,1))
        digital_feature = kornia.utils.image._to_bchw(torch.unsqueeze(digital_feature, 1))         
        digital_feature = torch.squeeze(kornia.morphology.closing(digital_feature, kernel))
        digital_feature = torch.permute(digital_feature, (1,2,0))

    return torch.squeeze(digital_feature)
